/*
 * Ray tracer using CUDA
 *
 * Uses code from nVidia's CUDA SDK Mandelbrot project 
 * (OpenGl/CUDA communication, OpenGL drawing, etc)
 * 
 * Author Mike Allgyer
*/

#include <stdlib.h> 
#include <stdio.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_runtime_api.h>
#include <cuda_gl_interop.h>
#include <cutil.h>

#include "fbx/Common.h"
#include "raytracer_kernel.h"
#include "Camera.h"
#include "ImportedModel.h"
#include "Util.h"

#define SCENE_FILE "scenes\\triangles.txt"
//openGL names
GLuint glPBO, glTex;

// host side image
uchar4 *hostImg;

// image dimensions
const int WIDTH = 800, HEIGHT = 600;

// timer 
unsigned int gTimer;

// frame rate 
float gFPS;

// the scene
ImportedModel *gImportedModel = 0;

// scene's camera
Camera *gCamera;

// dt for each frame
float dt = 0;

// how deep into into the hierarchy to descend
// (decides whether to draw bounding volumes or polygons
int gDrawDepth = MAX_DEPTH;

float3 gLight;

// instead of allocating/deallocating each
// frame, do it once, since objects don't move
int bvhCount = 0;

// bounding volumes
BoundingVolume **bvh = 0;

// recursively copy hierarchy
BoundingVolume *copyVolume(BoundingVolume *src) {
	BoundingVolume *dest = new BoundingVolume();
	BoundingVolume *hVolume = 0;
	BoundingVolume *child1 = NULL, *child2 = NULL;
	if (src->mNumVolumes > 0) { // add children first
		child1 = copyVolume(src->mVolume1);
		child2 = copyVolume(src->mVolume2);
	}

	dest->mVolume1 = child1;
	dest->mVolume2 = child2;

	dest->mCenterX = src->mCenterX;
	dest->mCenterY = src->mCenterY;
	dest->mCenterZ = src->mCenterZ;
	dest->mR = src->mR;
	dest->mG = src->mG;
	dest->mB = src->mB;
	dest->mReflection = src->mReflection;
	dest->mRefraction = src->mRefraction;
	dest->mIndexOfRefraction = src->mIndexOfRefraction;
	dest->mRadius = src->mRadius;
	dest->mNumPolygons = src->mNumPolygons;
	dest->mNumVolumes = src->mNumVolumes;

	// copy matrix
	Matrix *tempMatrix = NULL;
	if (src->mMatrix != NULL) {
		CUDA_SAFE_CALL(cudaMalloc((void**)&tempMatrix, sizeof(Matrix)));
		CUDA_SAFE_CALL(cudaMemcpy(tempMatrix, src->mMatrix, sizeof(Matrix), cudaMemcpyHostToDevice));
	}
	dest->mMatrix = tempMatrix;

	// copy polygons (and vertices)
	if (dest->mNumPolygons > 0) {
		// make a copy of polygon list
		Polygon *polys = new Polygon[src->mNumPolygons];
		for (int i = 0; i < src->mNumPolygons; i++) {
			polygonCopy(polys[i], src->mPolygons[i]);
			Vertex *v = 0;
			CUDA_SAFE_CALL(cudaMalloc((void**)&v, polys[i].vertexCount*sizeof(Vertex)));
			CUDA_SAFE_CALL(cudaMemcpy(v, polys[i].vertices, polys[i].vertexCount*sizeof(Vertex), cudaMemcpyHostToDevice));

			delete[] polys[i].vertices;
			polys[i].vertices = v;

		}

		Polygon *hPolys = 0;
		CUDA_SAFE_CALL(cudaMalloc((void**)&hPolys, src->mNumPolygons*sizeof(Polygon)));
		CUDA_SAFE_CALL(cudaMemcpy(hPolys, polys, src->mNumPolygons*sizeof(Polygon), cudaMemcpyHostToDevice));
		dest->mPolygons = hPolys;
	}

	CUDA_SAFE_CALL(cudaMalloc((void**)&hVolume, sizeof(BoundingVolume)));
	CUDA_SAFE_CALL(cudaMemcpy(hVolume, dest, sizeof(BoundingVolume), cudaMemcpyHostToDevice));

	return hVolume;
}


// copy entire bvh to device memory
BoundingVolume **copyBVH(BoundingVolume **src, int count) {

	BoundingVolume **dest = 0;
	BoundingVolume **temp = new BoundingVolume*[count];
	for (int i = 0; i < count; i++) {
		temp[i] = copyVolume(src[i]);
	}

	CUDA_SAFE_CALL(cudaMalloc((void**)&dest, (count)*sizeof(BoundingVolume*)));
	CUDA_SAFE_CALL(cudaMemcpy(dest, temp, (count)*sizeof(BoundingVolume*), cudaMemcpyHostToDevice));

	return dest;
}

// recursively free memory occupied by one volume
void freeVolume(BoundingVolume *volume) {
	// you have to copy to host to read any of it
	BoundingVolume *temp = (BoundingVolume *)malloc(sizeof(BoundingVolume));
	CUDA_SAFE_CALL(cudaMemcpy(temp, volume, sizeof(BoundingVolume), cudaMemcpyDeviceToHost));
	if (temp->mNumVolumes > 0) { // free children first
		freeVolume(temp->mVolume1);
		freeVolume(temp->mVolume2);
	}
	if (temp->mMatrix != NULL) {
		CUDA_SAFE_CALL(cudaFree(temp->mMatrix));
	}

	// free polygons
//	for (int i = 0; i < temp->mNumPolygons; i++) {
		Polygon *polys = (Polygon *)malloc(temp->mNumPolygons*sizeof(Polygon));
		CUDA_SAFE_CALL(cudaMemcpy(polys, temp->mPolygons, temp->mNumPolygons*sizeof(Polygon), cudaMemcpyDeviceToHost));
		
		// free vertices
		for (int j = 0; j < temp->mNumPolygons; j++) {
			CUDA_SAFE_CALL(cudaFree(polys[j].vertices));
		}
//	}
	CUDA_SAFE_CALL(cudaFree(temp->mPolygons));
	CUDA_SAFE_CALL(cudaFree(volume));
}

// free memory occupied by bvh
void freeBVH(BoundingVolume **bvh, int count) {
	BoundingVolume **localBvh = (BoundingVolume **)malloc(count*sizeof(BoundingVolume*));
	CUDA_SAFE_CALL(cudaMemcpy(localBvh, bvh, count*sizeof(BoundingVolume*), cudaMemcpyDeviceToHost));

	for (int i = 0; i < count; i++) {
		freeVolume((localBvh)[i]);
	}

	CUDA_SAFE_CALL(cudaFree(bvh));
}

//openGL display function
void draw() {
/*	for (unsigned int i = 0; i < gImportedModel->getNumGeometry(); i++) {
		Geometry *g = gImportedModel->getGeometry(i);
		vector<Mesh *> m = g->getMeshes();
		for (unsigned int j = 0; j < m.size(); j++) {
			m[j]->transformVertices();
		}
	}
	gImportedModel->recreateHierarchy();
*/
	uchar4 *dest;
	CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&dest, glPBO));


	// cuda work
	runRaytracer(dest, WIDTH, HEIGHT, gCamera->getPos(), gCamera->getLookAt(), gCamera->getUp(), 
		bvhCount, bvh, gImportedModel->getNumberOfLeaves(), gDrawDepth, gLight);
	
	CUDA_SAFE_CALL(cudaGLUnmapBufferObject(glPBO));

	// draw it
	glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, WIDTH, HEIGHT, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	glBegin(GL_TRIANGLES);
	glTexCoord2f(0.0f, 0.0f);
	glVertex2f(-1.0f, -1.0f);

	glTexCoord2f(2.0f, 0.0f);
	glVertex2f(3.0f, -1.0f);

	glTexCoord2f(0.0f, 2.0f);
	glVertex2f(-1.0f, 3.0f);

	glEnd();
	glutSwapBuffers();

	static int frameCount = 0;
	static int timeElapsed = 0;
	// get frame count
	++frameCount;
	int currTime = cutGetTimerValue(gTimer);
	if (currTime > timeElapsed) {
			dt = ((float)currTime - (float)timeElapsed) / 1000.0f;
	}
	timeElapsed = currTime;

	// calculate once a second
	if (currTime >= 1000) {
		gFPS = ((float)frameCount) / ((float)currTime / 1000.0f);
		frameCount = 0;
		cutResetTimer(gTimer);
		timeElapsed = timeElapsed - currTime;
	}


}

// openGL keyboard function
void keyboardFunc(unsigned char key, int, int) {
	// process keyboard here
	switch (key) {
		case 'q':
		case 'Q':
			// clean up

			CUT_SAFE_CALL(cutStopTimer(gTimer));
			CUT_SAFE_CALL(cutDeleteTimer(gTimer));
			CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(glPBO));
			freeBVH(bvh, bvhCount);
			glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
			glDeleteBuffers(1, &glPBO);
			glDeleteTextures(1, &glTex);
			free(hostImg);
			delete gCamera;
			delete gImportedModel;
			exit(0);
			break;
		case 'f':
		case 'F': printf("FPS: %.5f (%.2f sec)\n", gFPS, dt);
			break;
		case '1': gDrawDepth = 1;
			break;
		case '2': gDrawDepth = 2;
			break;
		case '3': gDrawDepth = 3;
			break;
		case '4': gDrawDepth = 4;
			break;
		case '5': gDrawDepth = 5;
			break;
		case '6': gDrawDepth = 6;
			break;
		case '7': gDrawDepth = 7;
			break;
		case '8': gDrawDepth = 8;
			break;
		case '9': gDrawDepth = 9;
			break;
		case '0': gDrawDepth = MAX_DEPTH;
			break;
		case 'w': 
		case 'W': gLight.z += 0.5f;
			break;
		case 's': 
		case 'S': gLight.z -= 0.5f;
			break;
		case 'a': 
		case 'A': gLight.x -= 0.5f;
			break;
		case 'd': 
		case 'D': gLight.x += 0.5f;
			break;
		case 'o': 
		case 'O': gCamera->rotateY(0.05f);
			break;
		case 'p': 
		case 'P': gCamera->rotateY(-0.05f);
			break;
		case 'i': 
		case 'I': gCamera->pitch(0.5f);
			break;
		case 'k': 
		case 'K': gCamera->pitch(-0.5f);
			break;
	}
}
// openGL 'special' keyboard function
void specialKeyboardFunc(int key, int, int) {
	// process keyboard here
	switch (key) {
		case GLUT_KEY_LEFT: gCamera->moveSideways(-1.8f);
			break;
		case GLUT_KEY_RIGHT: gCamera->moveSideways(1.8f);
			break;
		case GLUT_KEY_UP: gCamera->moveForward(1.8f);
			break;
		case GLUT_KEY_DOWN: gCamera->moveForward(-1.8f);
			break;
	}
}

// openGL click function
void mouseClick(int button, int, int x, int y) {
	// process clicks here
}

// openGL mouse movement function
void mouseMove(int x, int y) {
	// process mouse movement here
}

// openGL idle function
void idleFunc() {

	glutPostRedisplay();
}

// initialize CUDA and OpenGL 
int main(int argc, char **argv) {
	// initialize CUDA
	CUT_DEVICE_INIT();

	hostImg = (uchar4*)malloc(WIDTH*HEIGHT*4);


	glTranslatef(0,0,0);
	printf("Initializing GLUT...");

	// create window
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(WIDTH, HEIGHT);
	glutInitWindowPosition(256, 256);
	glutCreateWindow("Real-Time Ray Tracer Using CUDA");
	printf("Done.\n");


	printf("Loading extensions: %s\n", glewGetErrorString(glewInit()));
	if (!glewIsSupported(
		"GL_VERSION_2_0 " 
           "GL_ARB_pixel_buffer_object "
           "GL_EXT_framebuffer_object "
		   )) {
			   printf("Failed! (Required OpenGL Extensions not supported)\n");
			   CUT_EXIT(argc, argv);
			   return CUTFalse;
	}

	
	// create texture
	printf("Creating OpenGL Texture...");
	glEnable(GL_TEXTURE_2D);
	glGenTextures(1, &glTex);
	glBindTexture(GL_TEXTURE_2D, glTex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, WIDTH, HEIGHT, 0, GL_RGBA, GL_UNSIGNED_BYTE, hostImg);
	printf("Done.\n");

	printf("Creating Buffer...");
	// create buffer "PBO"
	glGenBuffers(1, &glPBO);
	glBindBuffer(GL_PIXEL_UNPACK_BUFFER_ARB, glPBO);
	glBufferData(GL_PIXEL_UNPACK_BUFFER_ARB, WIDTH * HEIGHT * 4, hostImg, GL_STREAM_COPY);

	CUDA_SAFE_CALL(cudaGLRegisterBufferObject(glPBO));
	printf("Done.\n");

	glutDisplayFunc(draw);
	glutIdleFunc(idleFunc);
	glutKeyboardFunc(keyboardFunc);
	glutSpecialFunc(specialKeyboardFunc);
	glutMouseFunc(mouseClick);
	glutMotionFunc(mouseMove);


	// start timer
	CUT_SAFE_CALL(cutCreateTimer(&gTimer));
	CUT_SAFE_CALL(cutStartTimer(gTimer));


	gLight.x = -5;
	gLight.y = 5;
	gLight.z = -5;

	// create scene
	gImportedModel = new ImportedModel(SCENE_FILE);

	gImportedModel->printBVH();

	// create camera
	gCamera = new Camera();

	BoundingVolume **volumes = gImportedModel->getHierarchies();
	bvhCount = gImportedModel->getBVHCount();
	bvh = copyBVH(volumes, bvhCount);

	glutMainLoop();

	CUT_EXIT(argc, argv);

}
