#include <cutil.h>
#include <stdio.h>
#include "raytracer_kernel.h"
#include "Vertex.h"
#include "Util.h"
//#include "BoundingVolume.h"

// The dimensions of the thread block
#define BLOCKDIM_X 8
#define BLOCKDIM_Y 8

// number of packets each thread computes
#define RAY_PACKET_SIZEX 4
#define RAY_PACKET_SIZEY 4

#define NO_INTERSECTION 0.0f
#define ERROR_AMT 0.01f
#define CAM_DIST 300.0f
#define KD_DEFAULT 0.8f
#define KS_DEFAULT 0.6f
#define KS_EXP 9.0f

#define RECURSION_DEPTH 4

#define PI 3.14159f

// used in reflection/refraction iteratons
struct Ray {
	float3 dir;
	float3 pos;
	float weight;
	float index;

//	__device__ Ray() {};
};

// allow for small error
__device__ inline bool equalsZero(float x) {
	return x <= ERROR_AMT && x >= -ERROR_AMT;
}
// allow for small error
__device__ inline bool equalTo(float x, float y) {
	return x <= (y + ERROR_AMT) && x >= (y - ERROR_AMT);
}
// get an angle between two vectors
__device__ inline float getAngle(float3 a, float3 b){
	float angle = a.x * b.x + a.y * b.y + a.z * b.z;
	return acos(angle);

}
// dot product
__device__ inline float dotProduct(float3 *r1, float3 *r2) {
	float dot = (r1->x*r2->x) +(r1->y*r2->y) +(r1->z*r2->z);
	return dot;
}

// cross product
__device__ inline float3 crossProduct(float3 r1, float3 r2) {
	float3 c;
	c.x = r1.z*r2.y - r1.y*r2.z;
	c.y = r1.x*r2.z - r1.z*r2.x;
	c.z = r1.y*r2.x - r1.x*r2.y;

	return c;
}

// decides wether a ray should be reversed in order to "face" 
// a certain direction
__device__ inline bool reverseRay(float3 ray, float3 revRay) {
	float diff = 0, diff2 = 0;
	diff += abs(ray.x - revRay.x);
	diff += abs(ray.y - revRay.y);
	diff += abs(ray.z - revRay.z);

	diff2 += abs(ray.x + revRay.x);
	diff2 += abs(ray.y + revRay.y);
	diff2 += abs(ray.z + revRay.z);

	return diff < diff2;
}
// normalizes a ray
__device__ inline void normalizeRay(float3 *dir) {
	float root = dir->x*dir->x + dir->y*dir->y + dir->z*dir->z;
	if (!(root == 0)) {

		// div since mult is faster
		float length = 1 / sqrtf(root);

		dir->x = dir->x * length;
		dir->y = dir->y * length;
		dir->z = dir->z * length;
	}
}

// get reflection ray
__device__ inline float3 getReflection(float3 *lightRay, float3 *nRay, float dot) {
	float3 ref;
	ref.x = lightRay->x - 2.0f * nRay->x * dot;
	ref.y = lightRay->y - 2.0f * nRay->y * dot;
	ref.z = lightRay->z - 2.0f * nRay->z * dot;
	return ref;
}

// performs ray-polygon intersection
// assumes ray is normalized
__device__ inline float getPolygonIntersection(float3 rayPos, float3 rayDir, Vertex verts[], float numVertices) {

	float3 normal;
	normal.x = verts[0].nx;
	normal.y = verts[0].ny;
	normal.z = verts[0].nz;

	float dist;

	normalizeRay(&normal);

	float denom = (normal.x*rayDir.x + normal.y*rayDir.y + normal.z*rayDir.z);
	if (denom >= -ERROR_AMT*0.25f && denom <= 0.25f*ERROR_AMT) {
		dist = NO_INTERSECTION;
	} else {
		float distPlaneOrigin = -1.0f * (normal.x*verts[0].x + normal.y*verts[0].y + normal.z*verts[0].z);
		float root = -1.0f * ((normal.x*rayPos.x + normal.y*rayPos.y + normal.z*rayPos.z + distPlaneOrigin) / denom);

		if (root <= ERROR_AMT) { // behind ray
			dist = NO_INTERSECTION;
		} else {
			dist = root;

			// is it inside polygon?
			// basic algorithm from http://www.siggraph.org/education/materials/HyperGraph/raytrace/raypolygon_intersection.htm
			float3 intPoint;
			intPoint.x = rayPos.x + rayDir.x * dist;
			intPoint.y = rayPos.y + rayDir.y * dist;
			intPoint.z = rayPos.z + rayDir.z * dist;

			// project polygon onto plane
			float dominate = max(abs(normal.x), max(abs(normal.y), abs(normal.z)));
			float2 testRay;
			if (dominate == abs(normal.x)) {
				testRay.x = intPoint.y;
				testRay.y = intPoint.z;
			} else if (dominate == abs(normal.y)) {
				testRay.x = intPoint.x;
				testRay.y = intPoint.z;
			} else {
				testRay.x = intPoint.x;
				testRay.y = intPoint.y;
			}
			// y = mx + b
			//float testM = testRay.y / testRay.x;
			//float testB = testRay.y - testM*testRay.x;
			int numHits = 0;
			for (int i = 0; i < numVertices; i++) {
				float2 v1, v2;
				int index = i+1;
				if (index >= numVertices) index = 0;
				if (dominate == abs(normal.x)) {
					v1.x = verts[i].y - testRay.x;
					v2.x = verts[index].y - testRay.x;
					v1.y = verts[i].z - testRay.y;
					v2.y = verts[index].z - testRay.y;
				} else if (dominate == abs(normal.y)) {
					v1.x = verts[i].x - testRay.x;
					v2.x = verts[index].x - testRay.x;
					v1.y = verts[i].z - testRay.y;
					v2.y = verts[index].z - testRay.y;
				} else {
					v1.x = verts[i].x - testRay.x;
					v2.x = verts[index].x - testRay.x;
					v1.y = verts[i].y - testRay.y;
					v2.y = verts[index].y - testRay.y;
				}
//v1.x=1;
//v1.y=1;
//v2.x=3;
//v2.y=-1;
				// not horizontal or vertical
				if (v1.x != v2.x && v1.y != v2.y) {

					float m = (v1.y - v2.y) / (v1.x - v2.x);
					float b = v1.y - m*v1.x;

					// get x-intercept
					float xInt = -b / m;
				if (xInt >= 0 && ((xInt >= v1.x && xInt <= v2.x) || (xInt >= v2.x && xInt <= v1.x))) ++numHits;

				} else if ((equalTo(v1.x, v2.x) && v1.x >= 0) && ((v1.y >= 0 && v2.y <= 0) || (v1.y <= 0 && v2.y >= 0))) {
					++numHits;	// vertical
				} else if (equalTo(v1.y, v2.y) && equalsZero(v1.y) && ((v1.x < 0 && v2.x > 0) || (v1.x > 0 && v2.x < 0))) {
//					++numHits;  // horizontal
				}
			}
			if ((numHits%2) == 0) {
				dist = NO_INTERSECTION;
			}
		// old way, gives noise
/*			float angle = 0.0f;
			for (int i = 0; i < numVertices; i++) {
				float3 r1;
				r1.x = intPoint.x - verts[i].x;
				r1.y = intPoint.y - verts[i].y;
				r1.z = intPoint.z - verts[i].z;
				normalizeRay(&r1);

				int index = (int)((i+1));
				if (index >= numVertices) index = 0;
				float3 r2;
				r2.x = intPoint.x - verts[index].x;
				r2.y = intPoint.y - verts[index].y;
				r2.z = intPoint.z - verts[index].z;
				normalizeRay(&r2);
					
				angle += getAngle(r1, r2);
			}
			if (angle > (2*PI + ERROR_AMT) || angle < (2*PI - ERROR_AMT)) {
				dist = NO_INTERSECTION;
			}
			*/
		}
	}

	return dist;
}
// performs ray-sphere intersection
// assumes ray is normalized
__device__ inline float getSphereIntersection(float3 *rayPos, float3 *rayDir, float3 sPos, float sRadius) {

	float retVal = 0.0f;
	float b, c, root, omega1, omega2;


	// not needed since ray is normalized
	//a = (dx * dx) + (dy * dy) + (dz * dz);

	b = 2 * (rayDir->x * (rayPos->x  - sPos.x) + rayDir->y * (rayPos->y - sPos.y) + rayDir->z * (rayPos->z - sPos.z));
	c = (rayPos->x - sPos.x) * (rayPos->x - sPos.x) + (rayPos->y - sPos.y) * (rayPos->y - sPos.y) + 
		(rayPos->z - sPos.z) * (rayPos->z - sPos.z) -	sRadius * sRadius;

	root = b*b - 4*c;

	// no intersection
	if (root < 0) {
		retVal = NO_INTERSECTION;
	}
	if (root == 0) {
		//retVal = -b + sqrtf(root)/2;
		retVal = -b;
	} 
	
	if (root > 0) {
		// sqrt is slow, and since this function is usually
		// used for comparison, do it later
		root = sqrtf(root);

		omega1 = (-b + root)/(2);
		omega2 = (-b - root)/(2);
		
		if ((omega1 < ERROR_AMT) && (omega2 < ERROR_AMT) ){  
			retVal = NO_INTERSECTION;
		} else if( (omega2 < ERROR_AMT || omega1 <= omega2) && (omega1 > ERROR_AMT)) {
			retVal = omega1;
		} else if( (omega1 < ERROR_AMT || omega2 <= omega1) && (omega2 > ERROR_AMT)){
			retVal = omega2;
		}
		
	}

	return retVal;
}

// gets the distance to the nearest object
// also gets the normal at point of intersection
__device__ void traverseWorld(float3 rayPos, float3 rayDir, BoundingVolume **bvh, int numSpheres, 
							  float3 &normalOut, float &intDist, float3 &color, float &reflection, float &refraction, 
							  float &indexOfRefraction, int maxLeaves, int drawDepth) {
	intDist = NO_INTERSECTION;
	float dist = NO_INTERSECTION;

	reflection = 0;
	refraction = 0;
	indexOfRefraction = 1.0f;

	BoundingVolume *volumes[50];
	BoundingVolume *children[50];	// this holds the volumes that 
									//contain children that need to be checked
	BoundingVolume *leaves[50];	// this holds the volumes that 
								//contain vertices that need to be checked

	for (int i = 0; i < numSpheres; i++) {

		// intialize 
		for (int j = 0; j < maxLeaves; j++) {
			volumes[j] = NULL;
			leaves[j] = NULL;
			children[j] = NULL;
		}

		// top-level volume
		volumes[0] = bvh[i];
		bool done = false;

		int currDepth = 0;

		// keep proceeding into hierarchy, saving children that contain hits
		while (!done && currDepth <= drawDepth) {
			currDepth++;
			int childIndex = 0;
			done = true;
			for (int i = 0; i < maxLeaves; i++) {
				if (volumes[i] != NULL) {
					float3 c;
					c.x = volumes[i]->mCenterX;
					c.y = volumes[i]->mCenterY;
					c.z = volumes[i]->mCenterZ;
					float d = getSphereIntersection(&rayPos, &rayDir, c, volumes[i]->mRadius);
					if (d != NO_INTERSECTION) {
						if (volumes[i]->mNumVolumes == 2 && currDepth < drawDepth) {
							// insert children into array
							children[childIndex] = volumes[i]->mVolume1;
							childIndex++;
							children[childIndex] = volumes[i]->mVolume2;
							childIndex++;

							done = false;
						} else { // move this to leaves
							for (int j = 0; j < maxLeaves; j++) {
								if (leaves[j] == NULL) {
									leaves[j] = volumes[i];
									break;
								}
							}
							volumes[i] = NULL;
						}
					}
				}
			}
			for (int j = 0; j < childIndex; j++) {
				volumes[j] = children[j];
			}
			for (int j = childIndex; j < (maxLeaves - childIndex); j++) {
				volumes[j] = NULL;
			}
			for (int j = 0; j < maxLeaves; j++) {
				children[j] = NULL;
			}
		}
		for (int j = 0; j < maxLeaves; j++) {
			if (leaves[j] != NULL) {
				// just draw spheres
				if (drawDepth != MAX_DEPTH || leaves[j]->mNumPolygons <= 0) {
					float3 c;
					c.x = leaves[j]->mCenterX;
					c.y = leaves[j]->mCenterY;
					c.z = leaves[j]->mCenterZ;
					dist =  getSphereIntersection(&rayPos, &rayDir, c, leaves[j]->mRadius);

					if (dist != NO_INTERSECTION && (intDist > dist || intDist == NO_INTERSECTION)) {
						// get  color and material properties of a polygon inside this volume
						BoundingVolume *temp = leaves[j];
						while (temp->mNumVolumes != 0 && temp->mNumPolygons == 0) {
							temp = temp->mVolume1;
						}
						if (temp->mNumPolygons != 0) {
							color.x = temp->mPolygons[0].r;
							color.y = temp->mPolygons[0].g;
							color.z = temp->mPolygons[0].b;
							reflection = temp->mPolygons[0].reflection;
							refraction = temp->mPolygons[0].refraction;
							indexOfRefraction = temp->mPolygons[0].indexOfRefraction;
						} else {  // use top-level volume
							color.x = leaves[j]->mR;
							color.y = leaves[j]->mG;
							color.z = leaves[j]->mB;
							reflection = leaves[j]->mReflection;
							refraction = leaves[j]->mRefraction;
							indexOfRefraction = leaves[j]->mIndexOfRefraction;
						}
						intDist = dist;
						float3 intPt;
						intPt.x = rayPos.x + rayDir.x*intDist;
						intPt.y = rayPos.y + rayDir.y*intDist;
						intPt.z = rayPos.z + rayDir.z*intDist;
						normalOut.x = intPt.x - leaves[j]->mCenterX;
						normalOut.y = intPt.y - leaves[j]->mCenterY;
						normalOut.z = intPt.z - leaves[j]->mCenterZ;
					}
				} else {
					// loop through polygons
					int numPolys = leaves[j]->mNumPolygons;
					dist = NO_INTERSECTION;
					bool isClose = false;
					int closePoly = 0;
					for (int k = 0; k < numPolys; k++) {
						dist = getPolygonIntersection(rayPos, rayDir, leaves[j]->mPolygons[k].vertices, 
							leaves[j]->mPolygons[k].vertexCount);

						if (dist != NO_INTERSECTION && (intDist > dist || intDist == NO_INTERSECTION)) {
							intDist = dist;
							normalOut.x = leaves[j]->mPolygons[k].vertices[0].nx;
							normalOut.y = leaves[j]->mPolygons[k].vertices[0].ny;
							normalOut.z = leaves[j]->mPolygons[k].vertices[0].nz;
							isClose = true;
							closePoly = k;
						}
					}
					if (isClose) {
						color.x = leaves[j]->mPolygons[closePoly].r;
						color.y = leaves[j]->mPolygons[closePoly].g;
						color.z = leaves[j]->mPolygons[closePoly].b;
						reflection = leaves[j]->mPolygons[closePoly].reflection;
						refraction = leaves[j]->mPolygons[closePoly].refraction;
						indexOfRefraction = leaves[j]->mPolygons[closePoly].indexOfRefraction;
					}
				}
			}
		}

//		delete[] volumes;
//		delete[] leaves;
		if (reverseRay(rayDir, normalOut)) {
			normalOut.x = -normalOut.x;
			normalOut.y = -normalOut.y;
			normalOut.z = -normalOut.z;
		}
		normalizeRay(&normalOut);
	}
}

/*
// copy single volume
void inline __device__ copyVolume(BoundingVolume dest, BoundingVolume *src) {
	dest.mCenterX = src->mCenterX;
	dest.mCenterY = src->mCenterY;
	dest.mCenterZ = src->mCenterZ;
	dest.mRadius = src->mRadius;
	dest.mNumPolygons = src->mNumPolygons;
	dest.mPolygons = src->mPolygons;
	dest.mR = src->mR;
	dest.mG = src->mG;
	dest.mB = src->mB;
	dest.mReflection = src->mReflection;
	dest.mRefraction = src->mRefraction;
	dest.mIndexOfRefraction = src->mIndexOfRefraction;
	dest.mNumVolumes = src->mNumVolumes;
	dest.mVolume1 = src->mVolume1;
	dest.mVolume2 = src->mVolume2;
	dest.mMatrix = src->mMatrix;

}

// copies an entire bvh to dest (usually shared memory)
__device__ inline void copyBVH(BoundingVolume* dest, BoundingVolume** src, int num) {
	BoundingVolume currV[50];
	BoundingVolume newV[50];
	int numCurr = num;
	int numNew = 0;
	
	// copy top-level volumes
	for (int i = 0; i < num; i++) {
		copyVolume(dest[i], src[i]);
		if (dest[i].mNumVolumes != 0) {
//			copyVolume(currV[numNew], src[i]->mVolume1);

		}
	}

	for (int i = 0; i < num; i++) {

	}
	
}
*/
// GPU threads that calculate pixels
__global__ void raytracer(uchar4 *dest, int width, int height, float3 camPos, float3 camLookAt, float3 camUp,
						  int color, int numSpheres, BoundingVolume **bvh, 
						  int maxLeaves, int drawDepth, float3 light) {

    const int x = (blockDim.x * blockIdx.x + threadIdx.x) * RAY_PACKET_SIZEX;
    const int y = (blockDim.y * blockIdx.y + threadIdx.y) * RAY_PACKET_SIZEY;
	const int stopX = x + RAY_PACKET_SIZEX;
	const int stopY = y + RAY_PACKET_SIZEY;

	// copy global memory to shared
/*#ifndef __DEVICE_EMULATION__
	// this works in CUDA, but gives emu* corrupted
	// stack error
	extern __shared__ float3 sCenters[];
	extern __shared__ float sRadii[];
#else
	float3 *sCenters = centers;
	float *sRadii = radii;
#endif

	for (int i = 0; i < numSpheres; i++) {
		sCenters[i] = centers[i];
		sRadii[i] = radii[i];
	}

*/
#ifndef __DEVICE_EMULATION__
	// this works in CUDA, but gives emu* corrupted
	// stack error
	extern __shared__ BoundingVolume* bvhh[];
//	copyBVH(bvhh, bvh, numSpheres);
#else
//	BoundingVolume bvh[numSpheres];
#endif

	int numLeaves = maxLeaves;

	for (int ix = x; ix < stopX; ix++) {
		for (int iy = y; iy < stopY; iy++) {
			if (ix < width && iy < height) {
				// y=0 on bottom
				int pixel = (height - iy - 1)*width + ix;
				float xInc = .1f;
				float yInc = .1f;

				float planeWidth = width * xInc;
				float planeHeight = height * yInc;

				// get this thread's ray

				// "center" pixel's ray:
				float3 rayDir;

				rayDir.x = (camLookAt.x - camPos.x);
				rayDir.y = (camLookAt.y - camPos.y);
				rayDir.z = (camLookAt.z - camPos.z);
				normalizeRay(&rayDir);

				rayDir.x *= CAM_DIST;
				rayDir.y *= CAM_DIST;
				rayDir.z *= CAM_DIST;

				// "left" direction
				normalizeRay(&camUp);
				float3 left = crossProduct(camUp, rayDir);
				normalizeRay(&left);
				float3 up = crossProduct(rayDir, left);
				normalizeRay(&up);

				left.x *= planeWidth/2.0f;
				left.y *= planeWidth/2.0f;
				left.z *= planeWidth/2.0f;

				up.x *= planeHeight/2.0f;
				up.y *= planeHeight/2.0f;
				up.z *= planeHeight/2.0f;

				normalizeRay(&left);
				normalizeRay(&up);
				// location of current pixel pixel
				float3 currPixel;
				currPixel.x = camPos.x + rayDir.x + (((float)width / 2 - ix) * xInc) * left.x + 
					(((float)height / 2 - iy) * yInc) * up.x;
				currPixel.y = camPos.y + rayDir.y + (((float)width / 2 - ix) * xInc) * left.y + 
					(((float)height / 2 - iy) * yInc) * up.y;
				currPixel.z = camPos.z + rayDir.z + (((float)width / 2 - ix) * xInc) * left.z + 
					(((float)height / 2 - iy) * yInc) * up.z;

				rayDir.x = (currPixel.x ); 
				rayDir.y = (currPixel.y ); 
				rayDir.z = (currPixel.z); 
				// normalize
				normalizeRay(&rayDir);

				float intDist = NO_INTERSECTION;
				float3 nRay;
				float reflection = 1, refraction = 0, indexOfRefraction = 1.0f;
				int count = 0;

				float3 finalColor;
				finalColor.x = 0;
				finalColor.y = 0;
				finalColor.z = 0;

#if RECURSION_DEPTH == 1
	#define MAX_RAYS 1
#elif  RECURSION_DEPTH == 2
	#define MAX_RAYS 4
#elif  RECURSION_DEPTH == 3
	#define MAX_RAYS 8
#elif  RECURSION_DEPTH == 4
	#define MAX_RAYS 16
#elif  RECURSION_DEPTH == 5
	#define MAX_RAYS 32
#else 
	#define MAX_RAYS 2
#endif
Ray currRays[MAX_RAYS];
Ray newRays[MAX_RAYS];

				int numRays = (int)pow(2.0f, RECURSION_DEPTH);
				currRays[0].pos = camPos;
				currRays[0].dir = rayDir;
				currRays[0].weight = 1.0f;
				currRays[0].index = 1.0f;
				int numCurrRays = 1;
				while (count < RECURSION_DEPTH && numCurrRays != 0) {
					++count;
					int numNewRays = 0;
					float3 rayPos;
					for (int currRay = 0; currRay < numCurrRays; currRay++) {
						rayPos = currRays[currRay].pos;
						rayDir = currRays[currRay].dir;
						float3 objColor;
						// default red
						objColor.x = .8f;
						objColor.y = .2f;
						objColor.z = .2f;

						float3 currColor;

						traverseWorld(rayPos, rayDir, bvh, numSpheres, nRay, intDist, objColor, reflection, 
						refraction, indexOfRefraction, numLeaves, drawDepth);
						
						if (intDist == NO_INTERSECTION) { // no intersection
							currColor.x = 155;
							currColor.y = 155;
							currColor.z = 200;
						} else {
							float3 intPoint;
							intPoint.x = rayPos.x + rayDir.x * intDist;
							intPoint.y = rayPos.y + rayDir.y * intDist;
							intPoint.z = rayPos.z + rayDir.z * intDist;

							// light ray
							float3 lightRay;
							lightRay.x = light.x - intPoint.x;
							lightRay.y = light.y - intPoint.y;
							lightRay.z = light.z - intPoint.z;
							normalizeRay(&lightRay);

							// shadows
							float3 normRay;
							float shadowDist = NO_INTERSECTION;
							float3 junkC;
							float junk;

							traverseWorld(intPoint, lightRay, bvh, numSpheres, normRay, shadowDist, junkC,
								junk, junk, junk, numLeaves, drawDepth);
							if (shadowDist != NO_INTERSECTION) {
//								reflection = 0;
//								refraction = 0;
								currColor.x = (int)(objColor.x*255.0f*.2f);
								currColor.y = (int)(objColor.y*255.0f*.2f);
								currColor.z = (int)(objColor.z*255.0f*.2f);
								currRays[currRay].weight *= 1.0f;
							} else {

								float diffuse = dotProduct(&lightRay, &nRay);
								diffuse = diffuse * .5f + .5f;

								// reflection ray
								float3 ref;
								float dot = dotProduct(&lightRay, &nRay);
								ref = getReflection(&lightRay, &nRay, dot);

								float specular = dotProduct(&ref, &rayDir);
								specular = pow(specular, KS_EXP);
								currColor.x = (((float)(objColor.x*255))*diffuse*KD_DEFAULT + 
									((float)(objColor.x*255))*specular*KS_DEFAULT);
								currColor.y = (((float)(objColor.y*255))*diffuse*KD_DEFAULT + 
									((float)(objColor.y*255))*specular*KS_DEFAULT);
								currColor.z = (((float)(objColor.z*255))*diffuse*KD_DEFAULT + 
									((float)(objColor.z*255))*specular*KS_DEFAULT);

							}
							if (reflection > 0) {
								float dot = dotProduct(&rayDir, &nRay);
								newRays[numNewRays].dir = getReflection(&rayDir, &nRay, dot);
								newRays[numNewRays].pos = intPoint;
								newRays[numNewRays].weight = currRays[currRay].weight*reflection;
								newRays[numNewRays].index = currRays[currRay].index;
								++numNewRays;
							}

							// refraction ray
							if (refraction > 0) {

								float nDotRay = dotProduct(&nRay, &rayDir);
								float3 tempNormal = nRay;

								float oldI = currRays[currRay].index;
								float newI = 1.0;
								if (nDotRay > 0) {
									tempNormal.x = -tempNormal.x;
									tempNormal.y = -tempNormal.y;
									tempNormal.z = -tempNormal.z;
									newI = indexOfRefraction;
								}

								nDotRay = dotProduct(&rayDir, &tempNormal);
								float root  = 1 - nDotRay*nDotRay;
								root *= oldI * oldI;
								root /= (newI * newI);
								root = 1 - root;

								if (root <= 0) { // total internal refraction
									float dot = dotProduct(&rayDir, &nRay);
									newRays[numNewRays].dir = getReflection(&rayDir, &nRay, dot);
									newRays[numNewRays].pos = intPoint;
									newRays[numNewRays].weight = currRays[currRay].weight*reflection;
									newRays[numNewRays].index = oldI;
									++numNewRays;
								} else {

									root = sqrt(root);
									float3 refRay1, refRay2;
									refRay1.x = tempNormal.x * root;
									refRay1.y = tempNormal.y * root;
									refRay1.z = tempNormal.z * root;
									normalizeRay(&refRay1);
									refRay2.x = tempNormal.x * nDotRay;
									refRay2.y = tempNormal.y * nDotRay;
									refRay2.z = tempNormal.z * nDotRay;
									normalizeRay(&refRay2);
									float nRatio = 1.0f / indexOfRefraction;
									refRay2.x = (rayDir.x - refRay2.x) * nRatio - refRay1.x;
									refRay2.y = (rayDir.y - refRay2.y) * nRatio - refRay1.y;
									refRay2.z = (rayDir.z - refRay2.z) * nRatio - refRay1.z;
									normalizeRay(&refRay2);
						 			newRays[numNewRays].dir = refRay2;
									newRays[numNewRays].pos = intPoint;
									newRays[numNewRays].weight = currRays[currRay].weight*refraction;
									newRays[numNewRays].index = newI;
									++numNewRays;
								}
							}
						}
						finalColor.x += currColor.x * currRays[currRay].weight;
						finalColor.y += currColor.y * currRays[currRay].weight;
						finalColor.z += currColor.z * currRays[currRay].weight;
					}
					// move new into current
					for (int newRay = 0; newRay < numNewRays; newRay++) {
						currRays[newRay].pos = newRays[newRay].pos;
						currRays[newRay].dir = newRays[newRay].dir;
						normalizeRay(&currRays[newRay].dir);
						currRays[newRay].weight = newRays[newRay].weight;
						currRays[newRay].index = newRays[newRay].index;
					}
					numCurrRays = numNewRays;
				}
				dest[pixel].x =(unsigned char)min(255.0f, finalColor.x);
				dest[pixel].y =(unsigned char)min(255.0f, finalColor.y);
				dest[pixel].z =(unsigned char)min(255.0f, finalColor.z);
			}
		}
	}
}

int pad(int a, int b) {
	return ((a % b) != 0) ? (a / b + 1) : (a / b);
}

// this spawns the GPU threads
void runRaytracer(uchar4 *dest, int width, int height, float3 camPos, float3 camLookAt, float3 camUp, 
				  int numSpheres, BoundingVolume **bvh, int maxLeaves, int drawDepth, float3 light) {
	dim3 threads(BLOCKDIM_X, BLOCKDIM_Y);
	dim3 grid(pad(pad(width, BLOCKDIM_X), RAY_PACKET_SIZEX), pad(pad(height, BLOCKDIM_Y), RAY_PACKET_SIZEY));

	static int color=150;
//	color=(color+1)%255;
	raytracer<<< grid, threads >>>(dest, width, height, camPos, camLookAt, camUp, 
		color, numSpheres, bvh, maxLeaves, drawDepth, light);

	cudaThreadSynchronize();
	CUT_CHECK_ERROR("CUDA Error\n");

} 